# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
import numpy as np
#
class bspline(object):
    """Bspline class.

    Functions in the bspline library are implemented as methods on this
    class.

    Parameters
    ----------
    x : :class:`numpy.ndarray`
        The data.
    nord : :class:`int`, optional
        To be documented.
    npoly : :class:`int`, optional
        To be documented.
    bkpt : :class:`numpy.ndarray`, optional
        To be documented.
    bkspread : :class:`float`, optional
        To be documented.
    verbose : :class:`bool`, optional.
        If ``True`` print extra information.

    Attributes
    ----------
    breakpoints
        To be documented.
    nord
        To be documented.
    npoly
        To be documented.
    mask
        To be documented.
    coeff
        To be documented.
    icoeff
        To be documented.
    xmin
        To be documented.
    xmax
        To be documented.
    funcname
        To be documented.
    """
    def __init__(self,x,nord=4,npoly=1,bkpt=None,bkspread=1.0,verbose=False,**kwargs):
        """Init creates an object whose attributes are similar to the
        structure returned by the create_bspline function.
        """
        #
        # Set the breakpoints.
        #
        if bkpt is None:
            startx = x.min()
            rangex = x.max() - startx
            if 'placed' in kwargs:
                w = ((kwargs['placed'] >= startx) &
                    (kwargs['placed'] <= startx+rangex))
                if w.sum() < 2:
                    bkpt = np.arange(2,dtype='f')*rangex + startx
                else:
                    bkpt = kwargs['placed'][w]
            elif 'bkspace' in kwargs:
                nbkpts = int(rangex/kwargs['bkspace']) + 1
                if nbkpts < 2:
                    nbkpts = 2
                tempbkspace = rangex/float(nbkpts-1)
                bkpt = np.arange(nbkpts,dtype='f')*tempbkspace + startx
            elif 'nbkpts' in kwargs:
                nbkpts = kwargs['nbkpts']
                if nbkpts < 2:
                    nbkpts = 2
                tempbkspace = rangex/float(nbkpts-1)
                bkpt = np.arange(nbkpts,dtype='f')*tempbkspace + startx
            elif 'everyn' in kwargs:
                npkpts = max(nx/kwargs['everyn'], 1)
                if nbkpts == 1:
                    xspot = [0]
                else:
                    xspot = int(nx/(nbkpts-1))*np.arange(nbkpts,dtype='i4')
                bkpt = x[xspot].astype('f')
            else:
                raise ValueError('No information for bkpts.')
        imin = bkpt.argmin()
        imax = bkpt.argmax()
        if x.min() < bkpt[imin]:
            if verbose:
                print('Lowest breakpoint does not cover lowest x value: changing.')
            bkpt[imin] = x.min()
        if x.max() > bkpt[imax]:
            if verbose:
                print('Highest breakpoint does not cover highest x value: changing.')
            bkpt[imax] = x.max()
        nshortbkpt = bkpt.size
        fullbkpt = bkpt.copy()
        if nshortbkpt == 1:
            bkspace = np.float32(bkspread)
        else:
            bkspace = (bkpt[1] - bkpt[0]) * np.float32(bkspread)
        for i in np.arange(1,nord,dtype=np.float32):
            fullbkpt = np.insert(fullbkpt,0,bkpt[0]-bkspace*i)
            fullbkpt = np.insert(fullbkpt,fullbkpt.shape[0],
                bkpt[nshortbkpt-1] + bkspace*i)
        #
        # Set the attributes
        #
        nc = fullbkpt.size - nord
        self.breakpoints = fullbkpt
        self.nord = nord
        self.npoly = npoly
        self.mask = np.ones((fullbkpt.size,),dtype='bool')
        if npoly > 1:
            self.coeff = np.zeros((npoly,nc),dtype='d')
            self.icoeff = np.zeros((npoly,nc),dtype='d')
        else:
            self.coeff = np.zeros((nc,),dtype='d')
            self.icoeff = np.zeros((nc,),dtype='d')
        self.xmin = 0.0
        self.xmax = 1.0
        self.funcname = 'legendre'
        return
    #
    #
    #
    def fit(self,xdata,ydata,invvar,x2=None):
        """Calculate a B-spline in the least-squares sense.

        Fit is based on two variables: x which is sorted and spans a large range
        where bkpts are required y which can be described with a low order
        polynomial.

        Parameters
        ----------
        xdata : :class:`numpy.ndarray`
            Independent variable.
        ydata : :class:`numpy.ndarray`
            Dependent variable.
        invvar : :class:`numpy.ndarray`
            Inverse variance of `ydata`.
        x2 : :class:`numpy.ndarray`, optional
            Orthogonal dependent variable for 2d fits.

        Returns
        -------
        fit : :func:`tuple`
            A tuple containing an integer error code, and the evaluation of the
            b-spline at the input values.  An error code of -2 is a failure,
            -1 indicates dropped breakpoints, 0 is success, and positive
            integers indicate ill-conditioned breakpoints.
        """
        from . import cholesky_band, cholesky_solve
        goodbk = self.mask[self.nord:]
        nn = goodbk.sum()
        if nn < self.nord:
            yfit = np.zeros(ydata.shape,dtype='f')
            return (-2,yfit)
        nfull = nn * self.npoly
        bw = self.npoly * self.nord
        a1,lower,upper = self.action(xdata,x2=x2)
        foo = np.tile(invvar,bw).reshape(bw,invvar.size).transpose()
        a2 = a1 * foo
        alpha = np.zeros((bw,nfull+bw),dtype='d')
        beta = np.zeros((nfull+bw,),dtype='d')
        bi = np.arange(bw,dtype='i4')
        bo = np.arange(bw,dtype='i4')
        for k in range(1,bw):
            bi = np.append(bi, np.arange(bw-k,dtype='i4')+(bw+1)*k)
            bo = np.append(bo, np.arange(bw-k,dtype='i4')+bw*k)
        for k in range(nn-self.nord+1):
            itop = k*self.npoly
            ibottom = min(itop,nfull) + bw - 1
            ict = upper[k] - lower[k] + 1
            if ict > 0:
                work = np.dot(a1[lower[k]:upper[k]+1,:].T,a2[lower[k]:upper[k]+1,:])
                wb = np.dot(ydata[lower[k]:upper[k]+1],a2[lower[k]:upper[k]+1,:])
                alpha.T.flat[bo+itop*bw] += work.flat[bi]
                beta[itop:ibottom+1] += wb
        min_influence = 1.0e-10 * invvar.sum() / nfull
        errb = cholesky_band(alpha,mininf=min_influence) # ,verbose=True)
        if isinstance(errb[0],int) and errb[0] == -1:
            a = errb[1]
        else:
            yfit,foo = self.value(xdata,x2=x2,action=a1,upper=upper,lower=lower)
            return (self.maskpoints(errb[0]),yfit)
        errs  = cholesky_solve(a,beta)
        if isinstance(errs[0],int) and errs[0] == -1:
            sol = errs[1]
        else:
            #
            # It is not possible for this to get called, because cholesky_solve
            # has only one return statement, & that statement guarantees that
            # errs[0] == -1
            #
            yfit,foo = self.value(xdata,x2=x2,action=a1,upper=upper,lower=lower)
            return (self.maskpoints(errs[0]),yfit)
        if self.npoly > 1:
            self.icoeff[:,goodbk] = np.array(a[0,0:nfull].reshape(self.npoly,nn),dtype=a.dtype)
            self.coeff[:,goodbk] = np.array(sol[0:nfull].reshape(self.npoly,nn),dtype=sol.dtype)
        else:
            self.icoeff[goodbk] = np.array(a[0,0:nfull],dtype=a.dtype)
            self.coeff[goodbk] = np.array(sol[0:nfull],dtype=sol.dtype)
        yfit,foo = self.value(xdata,x2=x2,action=a1,upper=upper,lower=lower)
        return (0,yfit)
    #
    #
    #
    def action(self,x,x2=None):
        """Construct banded bspline matrix, with dimensions [ndata, bandwidth].

        Parameters
        ----------
        x : :class:`numpy.ndarray`
            Independent variable.
        x2 : :class:`numpy.ndarray`, optional
            Orthogonal dependent variable for 2d fits.

        Returns
        -------
        action : :func:`tuple`
            A tuple containing the b-spline action matrix; the 'lower' parameter,
            a list of pixel positions, each corresponding to the first
            occurence of position greater than breakpoint indx; and 'upper',
            Same as lower, but denotes the upper pixel positions.
        """
        from ... import uniq
        from ...goddard.math import flegendre
        from ..trace import fchebyshev
        nx = x.size
        nbkpt = self.mask.sum()
        if nbkpt < 2*self.nord:
            return (-2,0,0)
        n = nbkpt - self.nord
        gb = self.breakpoints[self.mask]
        bw = self.npoly*self.nord
        lower = np.zeros((n - self.nord + 1,),dtype='i4')
        upper = np.zeros((n - self.nord + 1,),dtype='i4') -1
        indx = self.intrv(x)
        bf1 = self.bsplvn(x, indx)
        action = bf1
        aa = uniq(indx,np.arange(indx.size,dtype='i4'))
        upper[indx[aa]-self.nord+1] = aa
        rindx = indx[::-1]
        bb = uniq(rindx,np.arange(rindx.size,dtype='i4'))
        lower[rindx[bb]-self.nord+1] = nx - bb - 1
        if x2 is not None:
            if x2.size != nx:
                raise ValueError('Dimensions of x and x2 do not match.')
            x2norm = 2.0 * (x2 - self.xmin) / (self.xmax - self.xmin ) - 1.0
            if self.funcname == 'poly':
                temppoly = np.ones((nx,self.npoly),dtype='f')
                for i in range(1,self.npoly):
                    temppoly[:,i] = temppoly[:,i-1] * x2norm
            elif self.funcname == 'poly1':
                temppoly = np.tile(x2norm,self.npoly).reshape(nx,self.npoly)
                for i in range(1,self.npoly):
                    temppoly[:,i] = temppoly[:,i-1] * x2norm
            elif self.funcname == 'chebyshev':
                temppoly = fchebyshev(x2norm,self.npoly)
            elif self.funcname == 'legendre':
                temppoly = flegendre(x2norm,self.npoly)
            else:
                raise ValueError('Unknown value of funcname.')
            action = np.zeros((nx,bw),dtype='d')
            counter = -1
            for ii in range(self.nord):
                for jj in range(self.npoly):
                    counter += 1
                    action[:,counter] = bf1[:,ii]*temppoly[:,jj]
        return (action,lower,upper)
    #
    #
    #
    def intrv(self,x):
        """Find the segment between breakpoints which contain each value in the array x.

        The minimum breakpoint is nbkptord -1, and the maximum
        is nbkpt - nbkptord - 1.

        Parameters
        ----------
        x : :class:`numpy.ndarray`
            Data values, assumed to be monotonically increasing.

        Returns
        -------
        intrv : :class:`numpy.ndarray`
            Position of array elements with respect to breakpoints.
        """
        gb = self.breakpoints[self.mask]
        n = gb.size - self.nord
        indx = np.zeros((x.size,),dtype='i4')
        ileft = self.nord -1
        for i in range(x.size):
            while x[i] > gb[ileft+1] and ileft < n - 1:
                ileft += 1
            indx[i] = ileft
        return indx
    #
    #
    #
    def bsplvn(self,x,ileft):
        """To be documented.

        Parameters
        ----------
        x : :class:`numpy.ndarray`
            To be documented.
        ileft : :class:`int`
            To be documented

        Returns
        -------
        bsplvn : :class:`numpy.ndarray`
            To be documented.
        """
        bkpt = self.breakpoints[self.mask]
        vnikx = np.zeros((x.size,self.nord),dtype=x.dtype)
        deltap = vnikx.copy()
        deltam = vnikx.copy()
        j = 0
        vnikx[:,0] = 1.0
        while j < self.nord - 1:
            ipj = ileft+j+1
            deltap[:,j] = bkpt[ipj] - x
            imj = ileft-j
            deltam[:,j] = x - bkpt[imj]
            vmprev = 0.0
            for l in range(j+1):
                vm = vnikx[:,l]/(deltap[:,l] + deltam[:,j-l])
                vnikx[:,l] = vm*deltap[:,l] + vmprev
                vmprev = vm*deltam[:,j-l]
            j += 1
            vnikx[:,j] = vmprev
        return vnikx
    #
    #
    #
    def value(self,x,x2=None,action=None,lower=None,upper=None):
        """Evaluate a bspline at specified values.

        Parameters
        ----------
        x : :class:`numpy.ndarray`
            Independent variable.
        x2 : :class:`numpy.ndarray`, optional
            Orthogonal dependent variable for 2d fits.
        action : :class:`numpy.ndarray`, optional
            Action matrix to use.  If not supplied it is calculated.
        lower : :class:`numpy.ndarray`, optional
            If the action parameter is supplied, this parameter must also
            be supplied.
        upper : :class:`numpy.ndarray`, optional
            If the action parameter is supplied, this parameter must also
            be supplied.

        Returns
        -------
        value : :func:`tuple`
            A tuple containing the results of the bspline evaluation and a
            mask indicating where the evaluation was good.
        """
        xsort = x.argsort()
        xwork = x[xsort]
        if x2 is not None:
            x2work = x2[xsort]
        else:
            x2work = None
        if action is not None:
            if lower is None or upper is None:
                raise ValueError('Must specify lower and upper if action is set.')
        else:
            action,lower,upper = self.action(xwork,x2=x2work)
        yfit = np.zeros(x.shape,dtype=x.dtype)
        bw = self.npoly * self.nord
        spot = np.arange(bw,dtype='i4')
        goodbk = self.mask.nonzero()[0]
        coeffbk = self.mask[self.nord:].nonzero()[0]
        n = self.mask.sum() - self.nord
        if self.npoly > 1:
            goodcoeff = self.coeff[:,coeffbk]
        else:
            goodcoeff = self.coeff[coeffbk]
        # maskthis = np.zeros(xwork.shape,dtype=xwork.dtype)
        for i in range(n-self.nord+1):
            ict = upper[i] - lower[i] + 1
            if ict > 0:
                yfit[lower[i]:upper[i]+1] = np.dot(
                    action[lower[i]:upper[i]+1,:],goodcoeff[i*self.npoly+spot])
        yy = yfit.copy()
        yy[xsort] = yfit
        mask = np.ones(x.shape,dtype='bool')
        gb = self.breakpoints[goodbk]
        outside = ((x < gb[self.nord-1]) |
            (x > gb[n]))
        if outside.any():
            mask[outside] = False
        hmm = ((np.diff(goodbk) > 2).nonzero())[0]
        for jj in range(hmm.size):
            inside = ((x >= self.breakpoints[goodbk[hmm[jj]]]) &
                (x <= self.breakpoints[goodbk[hmm[jj]+1]-1]))
            if inside.any():
                mask[inside] = False
        return (yy,mask)
    #
    #
    #
    def maskpoints(self,err):
        """Perform simple logic of which breakpoints to mask.

        Parameters
        ----------
        err : :class:`numpy.ndarray`
            The list of indexes returned by the cholesky routines.

        Returns
        -------
        maskpoints : :class:`int`
            An integer indicating the results of the masking.  -1 indicates
            that the error points were successfully masked.  -2 indicates
            failure; the calculation should be aborted.

        Notes
        -----
        The mask attribute is modified, assuming it is possible to create the
        mask.
        """
        nbkpt = self.mask.sum()
        if nbkpt <= 2*self.nord:
            return -2
        hmm = err[np.unique(err/self.npoly)]/self.npoly
        n = nbkpt - self.nord
        if np.any(hmm >= n):
            return -2
        test = np.zeros(nbkpt,dtype='bool')
        for jj in range(-np.ceil(nord/2.0),nord/2.0):
            foo = np.where((hmm+jj) > 0,hmm+jj,np.zeros(hmm.shape,dtype=hmm.dtype))
            inside = np.where((foo+nord) < n-1,foo+nord,np.zeros(hmm.shape,dtype=hmm.dtype)+n-1)
            test[inside] = True
        if test.any():
            reality = self.mask[test]
            if self.mask[reality].any():
                self.mask[reality] = False
                return -1
            else:
                return -2
        else:
            return -2



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
import numpy as np
from astropy.io.fits.fitsrec import FITS_rec
from . import fpoly, fchebyshev, func_fit
from .. import PydlutilsException
from ..math import djs_reject
from ..misc import djs_laxisgen
from ...goddard.math import flegendre
#
class TraceSet(object):
    """Implements the idea of a trace set.

    Attributes
    ----------
    func : str
        Name of function type used to fit the trace set.
    xmin : float-like
        Minimum x value.
    xmax : float-like
        Maximum x value.
    coeff : array-like
        Coefficients of the trace set fit.
    nTrace : int
        Number of traces in the object.
    ncoeff : int
        Number of coefficients of the trace set fit.
    xjumplo : float-like
        Jump value, for BOSS readouts.
    xjumphi : float-like
        Jump value, for BOSS readouts.
    xjumpval : float-like
        Jump value, for BOSS readouts.
    outmask : array-like.
        When initialized with x,y positions, this contains the rejected points.
    yfit : array-like.
        When initialized with x,y positions, this contains the fitted y values.
    """
    _func_map = {'poly':fpoly,'legendre':flegendre,'chebyshev':fchebyshev}
    #
    #
    #
    def __init__(self,*args,**kwargs):
        """This class can be initialized either with a set of xy positions,
        or with a trace set HDU from a FITS file.
        """
        if len(args) == 1 and isinstance(args[0],FITS_rec):
            #
            # Initialize with FITS data
            #
            self.func = args[0]['FUNC'][0]
            self.xmin = args[0]['XMIN'][0]
            self.xmax = args[0]['XMAX'][0]
            self.coeff = args[0]['COEFF'][0]
            self.nTrace = self.coeff.shape[0]
            self.ncoeff = self.coeff.shape[1]
            if 'XJUMPLO' in args[0].dtype.names:
                self.xjumplo = args[0]['XJUMPLO'][0]
                self.xjumphi = args[0]['XJUMPHI'][0]
                self.xjumpval = args[0]['XJUMPVAL'][0]
            else:
                self.xjumplo = None
                self.xjumphi = None
                self.xjumpval = None
            self.outmask = None
            self.yfit = None
        elif len(args) == 2:
            #
            # Initialize with x, y positions.
            #
            xpos = args[0]
            ypos = args[1]
            self.nTrace = xpos.shape[0]
            if 'invvar' in kwargs:
                invvar = kwargs['invvar']
            else:
                invvar = np.ones(xpos.shape,dtype=xpos.dtype)
            if 'func' in kwargs:
                self.func = kwargs['func']
            else:
                self.func = 'legendre'
            if 'ncoeff' in kwargs:
                self.ncoeff = int(kwargs['ncoeff'])
            else:
                self.ncoeff = 3
            if 'xmin' in kwargs:
                self.xmin = np.float64(kwargs['xmin'])
            else:
                self.xmin = xpos.min()
            if 'xmax' in kwargs:
                self.xmax = np.float64(kwargs['xmax'])
            else:
                self.xmax = xpos.max()
            if 'maxiter' in kwargs:
                maxiter = int(kwargs['maxiter'])
            else:
                maxiter = 10
            if 'inmask' in kwargs:
                inmask = kwargs['inmask']
            else:
                inmask = np.ones(xpos.shape,dtype=np.bool)
            do_jump = False
            if 'xjumplo' in kwargs:
                do_jump = True
                self.xjumplo = np.float64(kwargs['xjumplo'])
            else:
                self.xjumplo = None
            if 'xjumphi' in kwargs:
                self.xjumphi = np.float64(kwargs['xjumphi'])
            else:
                self.xjumphi = None
            if 'xjumpval' in kwargs:
                self.xjumpval = np.float64(kwargs['xjumpval'])
            else:
                self.xjumpval = None
            self.coeff = np.zeros((self.nTrace,self.ncoeff),dtype=xpos.dtype)
            self.outmask = np.zeros(xpos.shape,dtype=np.bool)
            self.yfit = np.zeros(xpos.shape,dtype=xpos.dtype)
            for iTrace in range(self.nTrace):
                xvec = self.xnorm(xpos[iTrace,:],do_jump)
                iIter = 0
                qdone = False
                tempivar = invvar[iTrace,:] * inmask[iTrace,:].astype(invvar.dtype)
                thismask = tempivar > 0
                while (not qdone) and (iIter <= maxiter):
                    res,ycurfit = func_fit(xvec, ypos[iTrace,:], self.ncoeff,
                        invvar=tempivar, function_name=self.func)
                    thismask,qdone = djs_reject(ypos[iTrace,:],ycurfit,invvar=tempivar)
                    iIter += 1
                self.yfit[iTrace,:] = ycurfit
                self.coeff[iTrace,:] = res
                self.outmask[iTrace,:] = thismask
        else:
            raise PydlutilsException("Wrong number of arguments to TraceSet!")
    #
    #
    #
    def xy(self,xpos=None,ignore_jump=False):
        """Convert from a trace set to an array of x,y positions.

        Parameters
        ----------
        xpos : array-like, optional
            If provided, evaluate the trace set at these positions.  Otherwise
            the positions will be constructed from the trace set object iself.
        ignore_jump : bool, optional
            If ``True``, ignore any jump information in the `tset` object

        Returns
        -------
        xy : tuple of array-like
            The x, y positions.
        """
        do_jump = self.has_jump and (not ignore_jump)
        if xpos is None:
            xpos = djs_laxisgen([self.nTrace,self.nx], iaxis=1) + self.xmin
        ypos = np.zeros(xpos.shape,dtype=xpos.dtype)
        for iTrace in range(self.nTrace):
            xvec = self.xnorm(xpos[iTrace,:],do_jump)
            legarr = self._func_map[self.func](xvec,self.ncoeff)
            ypos[iTrace:] = np.dot(legarr.T,self.coeff[iTrace,:])
        return (xpos,ypos)
    #
    #
    #
    @property
    def has_jump(self):
        return self.xjumplo is not None
    #
    #
    #
    @property
    def xRange(self):
        return self.xmax - self.xmin
    #
    #
    #
    @property
    def nx(self):
        return int(self.xRange + 1)
    #
    #
    #
    @property
    def xmid(self):
        return 0.5 * (self.xmin + self.xmax)
    #
    #
    #
    def xnorm(self,xinput,jump):
        """Convert input x coordinates to normalized coordinates suitable
        for input to special polynomials.
        """
        if jump:
            # Vector specifying what fraction of the jump has passed:
            jfrac = np.minimum(np.maximum(((xinput - self.xjumplo) / (self.xjumphi - self.xjumplo)),0.),1.)
            # Conversion to "natural" x baseline:
            xnatural = xinput + jfrac * self.xjumpval
        else:
            xnatural = xinput
        return 2.0 * (xnatural-self.xmid)/self.xRange



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_gcirc():
    from .. import gcirc
    import numpy as np
    from astropy.tests.helper import raises
    np.random.seed(137)
    #
    # Start in radians
    #
    offset = 5.0e-6 # approx 1 arcsec
    ra1 = 2.0*np.pi*np.random.rand(100)
    dec1 = np.pi/2.0 - np.arccos(2.0*np.random.rand(100) - 1.0)
    ra2 = ra1 + offset
    ra2 = np.where((ra2 > 2.0*np.pi), ra2 - 2.0*np.pi, ra2)
    dec2 = np.where((dec1 > 0), dec1 - offset, dec1 + offset)
    deldec2 = (dec2-dec1)/2.0
    delra2 =  (ra2-ra1)/2.0
    sindis = np.sqrt( np.sin(deldec2)*np.sin(deldec2) +
        np.cos(dec1)*np.cos(dec2)*np.sin(delra2)*np.sin(delra2) )
    dis = 2.0*np.arcsin(sindis)
    #
    # units = 0
    #
    d0 = gcirc(ra1,dec1,ra2,dec2,units=0)
    assert np.allclose(d0,dis)
    #
    # units = 2
    #
    d0 = gcirc(np.rad2deg(ra1)/15.0,np.rad2deg(dec1),np.rad2deg(ra2)/15.0,np.rad2deg(dec2),units=1)
    assert np.allclose(d0,np.rad2deg(dis)*3600.0)
    #
    # units = 2
    #
    d0 = gcirc(np.rad2deg(ra1),np.rad2deg(dec1),np.rad2deg(ra2),np.rad2deg(dec2),units=2)
    assert np.allclose(d0,np.rad2deg(dis)*3600.0)
    #
    # Units = whatever
    #
    with raises(ValueError):
        d0 = gcirc(ra1,dec1,ra2,dec2,units=5)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
#
from os import environ
#
class PathTestCase(object):
    """Set up environment variables for testing sdss_path and sdss_name.
    """
    path_envvars = ('PHOTO_CALIB','PHOTO_DATA','BOSS_PHOTOOBJ','PHOTO_REDUX',
        'PHOTO_RESOLVE','PHOTO_SKY','PHOTO_SWEEP')
    env_cache = {}
    path_data = {
        'apObj':"/PHOTO_REDUX/301/137/objcs/4",
        'calibMatch':"/PHOTO_REDUX/301/137/nfcalib",
        'calibPhotom':"/PHOTO_REDUX/301/137/nfcalib",
        'calibPhotomGlobal':"/PHOTO_CALIB/301/137/nfcalib",
        'fakeIdR':"/PHOTO_DATA/137/fake_fields/4",
        'fpAtlas':"/PHOTO_REDUX/301/137/objcs/4",
        'fpBIN':"/PHOTO_REDUX/301/137/objcs/4",
        'fpC':"/PHOTO_REDUX/301/137/objcs/4",
        'fpFieldStat':"/PHOTO_REDUX/301/137/objcs/4",
        'fpM':"/PHOTO_REDUX/301/137/objcs/4",
        'fpObjc':"/PHOTO_REDUX/301/137/objcs/4",
        'hoggObj':"/PHOTO_REDUX/301/137/objcs/4",
        'idFF':"/PHOTO_REDUX/301/137/objcs/4",
        'idR':"/PHOTO_DATA/137/fields/4",
        'idRR':"/PHOTO_DATA/137/fields/4",
        'psBB':"/PHOTO_REDUX/301/137/objcs/4",
        'psFF':"/PHOTO_REDUX/301/137/objcs/4",
        'psField':"/PHOTO_REDUX/301/137/objcs/4",
        'reObjGlobal':"/PHOTO_RESOLVE/301/137/resolve/4",
        'reObjRun':"/PHOTO_REDUX/301/137/resolve/4",
        'reObjTmp':"/PHOTO_RESOLVE/301/137/resolve/4",
        'tsField':"/PHOTO_REDUX/301/137/calibChunks/4",
        }
    name_data = {
        'apObj':"apObj-000137-r4-0042.fit",
        'calibMatch':"calibMatch-000137-4.fits",
        'calibPhotom':"calibPhotom-000137-4.fits",
        'calibPhotomGlobal':"calibPhotomGlobal-000137-4.fits",
        'fakeIdR':"idR-000137-r4-0042.fit",
        'fpAtlas':"fpAtlas-000137-4-0042.fit",
        'fpBIN':"fpBIN-000137-r4-0042.fit",
        'fpC':"fpC-000137-r4-0042.fit",
        'fpFieldStat':"fpFieldStat-000137-4-0042.fit",
        'fpM':"fpM-000137-r4-0042.fit",
        'fpObjc':"fpObjc-000137-4-0042.fit",
        'hoggObj':"hoggObj-000137-4-0042.fits",
        'idFF':"idFF-000137-r4.fit",
        'idR':"idR-000137-r4-0042.fit",
        'idRR':"idRR-000137-r4-0042.fit",
        'psBB':"psBB-000137-r4-0042.fit",
        'psFF':"psFF-000137-r4.fit",
        'psField':"psField-000137-4-0042.fit",
        'reObjGlobal':"reObjGlobal-000137-4-0042.fits",
        'reObjRun':"reObjRun-000137-4-0042.fits",
        'reObjTmp':"reObjTmp-000137-4-0042.fits",
        'tsField':"tsField-000137-4-301-0042.fit",
        }
    def setup(self):
        for e in self.path_envvars:
            try:
                actual_env = environ[e]
            except KeyError:
                actual_env = None
            self.env_cache[e] = actual_env
            environ[e] = '/'+e
        return
    def teardown(self):
        for e in self.env_cache:
            if self.env_cache[e] is None:
                del environ[e]
            else:
                environ[e] = self.env_cache[e]
        return



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def window_read(**kwargs):
    """Read window files in $PHOTO_RESOLVE.
    """
    from os import getenv
    from os.path import exists, join
    from . import window_score
    from .. import PhotoopException
    from ...pydlutils.mangle import set_use_caps
    from astropy.io import fits
    import numpy as np
    resolve_dir = getenv('PHOTO_RESOLVE')
    if resolve_dir is None:
        raise PhotoopException('You have not set the environment variable PHOTO_RESOLVE!')
    if 'silent' not in kwargs:
        kwargs['silent'] = True
    r = dict()
    if 'flist' in kwargs:
        if 'rescore' in kwargs:
            flist_file = join(resolve_dir,'window_flist_rescore.fits')
            if not exists(rescore_file):
                #
                # This will be called if window_flist_rescore.fits doesn't exist.
                #
                window_score()
        else:
            flist_file = join(resolve_dir,'window_flist.fits')
        with fits.open(rescore_file) as fit:
            r['flist'] = fit[1].data
    if 'blist' in kwargs or 'balkans' in kwargs:
        blist_file = join(resolve_dir,'window_blist.fits')
        with fits.open(balkan_file) as fit:
            r['blist'] = fit[1].data
    if 'bcaps' in kwargs or 'balkans' in kwargs:
        bcaps_file = join(resolve_dir,'window_bcaps.fits')
        with fits.open(bcaps_file) as fit:
            r['bcaps'] = fit[1].data
    if 'findx' in kwargs:
        findx_file = join(resolve_dir,'window_findx.fits')
        with fits.open(findx_file) as fit:
            r['findx'] = fit[1].data
    if 'bindx' in kwargs:
        bindx_file = join(resolve_dir,'window_bindx.fits')
        with fits.open(bindx_file) as fit:
            r['bindx'] = fit[1].data
    if 'balkans' in kwargs:
        #
        # Copy blist data to balkans
        #
        r['balkans'] = r['blist'].copy()
        r['balkans']['caps'] = { 'X':list(), 'CM':list() }
        r['balkans']['use_caps'] = np.zeros(r['balkans']['ICAP'].shape,dtype=np.uint64)
        if 'blist' not in kwargs:
            del r['blist']
        #
        # Copy bcaps data into balkans
        #
        for k in range(len(r['balkans']['ICAP'])):
            r['balkans']['caps']['X'].append(r['bcaps']['X'][balkans['ICAP'][k]:balkans['ICAP'][k]+balkans['NCAPS'][k]])
            r['balkans']['caps']['CM'].append(r['bcaps']['CM'][balkans['ICAP'][k]:balkans['ICAP'][k]+balkans['NCAPS'][k]])
            r['balkans']['use_caps'][k] = set_use_caps(
                r['balkans']['caps']['X'][k],
                r['balkans']['caps']['CM'][k],
                r['balkans']['use_caps'][k],
                allow_doubles=True)
        if 'bcaps' not in kwargs:
            del r['bcaps']
    return r



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
#
from astropy import units as u
from astropy.coordinates import (Angle, BaseCoordinateFrame, FrameAttribute,
    SphericalRepresentation, QuantityFrameAttribute,
    RepresentationMapping)
from .stripe_to_incl import stripe_to_incl
#
class SDSSMuNu(BaseCoordinateFrame):
    """SDSS Great Circle Coordinates

    Attributes
    ----------
    stripe
        SDSS Stripe Number
    node
        Node of the great circle with respect to the celestial equator.
        In SDSS, this is almost always RA = 95.0 degrees.
    incl
        Inclination of the great circle with respect to the celestial
        equator.
    phi
        Counter-clockwise position angle w.r.t. north for an arc
        in the +nu direction.

    Parameters
    ----------
    mu : :class:`~astropy.coordinates.Angle`
        Angle corresponding to longitude measured along a stripe.
    nu : :class:`~astropy.coordinates.Angle`
        Angle corresponding to latitude measured perpendicular to a stripe.

    Notes
    -----
    http://www.sdss.org/dr12/algorithms/surveycoords/
    """
    default_representation = SphericalRepresentation
    frame_specific_representation_info = {
        'spherical': [
            RepresentationMapping(reprname='lon', framename='mu', defaultunit=u.deg),
            RepresentationMapping(reprname='lat', framename='nu', defaultunit=u.deg)
            ]
        }
    frame_specific_representation_info['unitspherical'] = frame_specific_representation_info['spherical']
    stripe = FrameAttribute(default=0)
    node = QuantityFrameAttribute(default=Angle(95.0,unit=u.deg),unit=u.deg)
    # phi = QuantityFrameAttribute(default=None,unit=u.deg)
    #
    #
    #
    @property
    def incl(self):
        return Angle(stripe_to_incl(self.stripe),unit=u.deg)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_aesthetics():
    from .. import aesthetics
    from ... import Pydlspec2dException
    import numpy as np
    from astropy.tests.helper import raises
    np.random.seed(137)
    flux = np.random.rand(100)
    ivar = np.random.rand(100)
    #
    # No bad
    #
    f = aesthetics(flux,ivar)
    assert (f == flux).all()
    #
    # Bad points
    #
    ivar[ivar < 0.1] = 0.0
    #
    # Bad method
    #
    with raises(Pydlspec2dException):
        f = aesthetics(flux,ivar,'badmethod')
    #
    # Nothing
    #
    f = aesthetics(flux,ivar,'nothing')
    assert (f == flux).all()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def func_fit(x,y,ncoeff,invvar=None,function_name='legendre',ia=None,inputans=None,inputfunc=None):
    """Fit X, Y positions to a functional form.

    Parameters
    ----------
    x : array-like
        X values (independent variable).
    y : array-like
        Y values (dependent variable).
    ncoeff : int
        Number of coefficients to fit.
    invvar : array-like, optional
        Weight values; inverse variance.
    function_name : str, optional
        Function name, default 'legendre'.
    ia : array-like, optional
        An array of bool of length `ncoeff` specifying free (``True``) and
        fixed (``False``) parameters.
    inputans : array-like, optional
        An array of values of length `ncoeff` specifying the values of
        the fixed parameters.
    inputfunc : array-like, optional
        Multiply the function fit by these values.

    Returns
    -------
    func_fit : tuple of array-like
        Fit coefficients, length `ncoeff`; fitted values.
    """
    import numpy as np
    from . import fchebyshev, fchebyshev_split, fpoly
    from ...goddard.math import flegendre
    if x.shape != y.shape:
        raise ValueError('Dimensions of X and Y do not agree!')
    if invvar is None:
        invvar = np.ones(x.shape,dtype=x.dtype)
    else:
        if invvar.shape != x.shape:
            raise ValueError('Dimensions of X and invvar do not agree!')
    if ia is None:
        ia = np.ones((ncoeff,),dtype=np.bool)
    if not ia.all():
        if inputans is None:
            inputans = np.zeros((ncoeff,),dtype=x.dtype)
    #
    # Select unmasked points
    #
    igood = (invvar > 0).nonzero()[0]
    ngood = len(igood)
    res = np.zeros((ncoeff,),dtype=x.dtype)
    yfit = np.zeros(x.shape,dtype=x.dtype)
    if ngood == 0:
        pass
    elif ngood == 1:
        res[0] = y[igood[0]]
        yfit += y[igood[0]]
    else:
        ncfit = min(ngood,ncoeff)
        function_map = {
            'legendre':flegendre,
            'flegendre':flegendre,
            'chebyshev':fchebyshev,
            'fchebyshev':fchebyshev,
            'chebyshev_split':fchebyshev_split,
            'fchebyshev_split':fchebyshev_split,
            'poly':fpoly,
            'fpoly':fpoly
            }
        try:
            legarr = function_map[function_name](x,ncfit)
        except KeyError:
            raise KeyError('Unknown function type: {0}'.format(function_name))
        if inputfunc is not None:
            if inputfunc.shape != x.shape:
                raise ValueError('Dimensions of X and inputfunc do not agree!')
            legarr *= np.tile(inputfunc,ncfit).reshape(ncfit,x.shape[0])
        yfix = np.zeros(x.shape,dtype=x.dtype)
        nonfix = ia[0:ncfit].nonzero()[0]
        nparams = len(nonfix)
        fixed = (~ia[0:ncfit]).nonzero()[0]
        if len(fixed) > 0:
            yfix = np.dot(legarr.T,inputans * (1 - ia))
            ysub = y - yfix
            finalarr = legarr[nonfix,:]
        else:
            finalarr = legarr
            ysub = y
        # extra2 = finalarr * np.outer(np.ones((nparams,),dtype=x.dtype),(invvar > 0))
        extra2 = finalarr * np.outer(np.ones((nparams,),dtype=x.dtype),invvar)
        alpha = np.dot(finalarr,extra2.T)
        # assert alpha.dtype == x.dtype
        if nparams > 1:
            # beta = np.dot(ysub * (invvar > 0), finalarr.T)
            beta = np.dot(ysub * invvar, finalarr.T)
            assert beta.dtype == x.dtype
            # uu,ww,vv = np.linalg.svd(alpha,full_matrices=False)
            res[nonfix] = np.linalg.solve(alpha,beta)
        else:
            # res[nonfix] = (ysub * (invvar > 0) * finalarr).sum()/alpha
            res[nonfix] = (ysub * invvar * finalarr).sum()/alpha
        if len(fixed) > 0:
            res[fixed] = inputans[fixed]
        yfit = np.dot(legarr.T,res[0:ncfit])
    return (res,yfit)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the window directory of photoop.
"""
from .sdss_score import sdss_score
from .window_read import window_read
from .window_score import window_score



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
#
from .. import sdss_name
from os import environ
from os.path import join
from astropy.tests.helper import raises
from .PathTestCase import PathTestCase
#
class TestSdssName(PathTestCase):
    """Set of tests for sdss_path().
    """
    def test_sdss_name_bad_ftype(self):
        #
        # Bad ftype
        #
        with raises(KeyError):
            p = sdss_name('fooBar',137,4,42)
    def test_sdss_name(self):
        for ftype in self.name_data:
            assert sdss_name(ftype,137,4,42,'301','r',no_path=True) == self.name_data[ftype]
    def test_sdss_name_with_path(self):
        for ftype in self.name_data:
            assert sdss_name(ftype,137,4,42,'301','r') == join(self.path_data[ftype],self.name_data[ftype])
    def test_sdss_name_reObj(self):
        assert sdss_name('reObj',137,4,42,'301','r',no_path=True) == self.name_data['reObjGlobal']
        resolve = environ['PHOTO_RESOLVE']
        del environ['PHOTO_RESOLVE']
        assert sdss_name('reObj',137,4,42,'301','r',no_path=True) == self.name_data['reObjRun']
        environ['PHOTO_RESOLVE'] = resolve



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def filter_thru(flux,waveimg=None,wset=None,mask=None,filter_prefix='sdss_jun2001',toair=False):
    """Compute throughput in SDSS filters.

    Parameters
    ----------
    flux : array-like
        Spectral flux.
    waveimg : array-like, optional
        Full wavelength solution, with the same shape as `flux`.
    wset : :class:`TraceSet`, optional
        A trace set containing the wavelength solution.  Must be specified
        if `waveimg` is not specified.
    mask : array-like, optional
        Interpolate over pixels where `mask` is non-zero.
    filter_prefix : str, optional
        Specifies a set of filter curves.
    toair : bool, optional
        If ``True``, convert the wavelengths to air from vacuum before computing.

    Returns
    -------
    filter_thru : array-like
        Integrated flux in the filter bands.
    """
    from os import getenv
    from os.path import dirname,join
    import numpy as np
    from astropy.io import ascii
    from astropy.utils.data import get_pkg_data_filename
    from ...goddard.astro import vactoair
    from ...pydlutils.image import djs_maskinterp
    from ...pydlutils.trace import traceset2xy, xy2traceset
    nTrace,nx = flux.shape
    if filter_prefix != 'sdss_jun2001':
        raise ValueError("Filters other than {0} are not available!".format('sdss_jun2001'))
    ffiles = [(join(dirname(__file__),'..','..','pydlutils','data','filters','{0}_{1}_atm.dat'.format(filter_prefix,f))) for f in 'ugriz']
    if waveimg is None and wset is None:
        raise ValueError("Either waveimg or wset must be specified!")
    if waveimg is None:
        pixnorm, logwave = traceset2xy(wset)
        waveimg = 10**logwave
    if toair:
        newwaveimg = vactoair(waveimg)
    else:
        newwaveimg = waveimg
    logwave = np.log10(newwaveimg)
    diffx = np.outer(np.ones((nTrace,),dtype=flux.dtype),np.arange(nx-1,dtype=flux.dtype))
    diffy = logwave[:,1:] - logwave[:,0:nx-1]
    diffset = xy2traceset(diffx,diffy,ncoeff=4,xmin=0,xmax=nx-1)
    pixnorm, logdiff = traceset2xy(diffset)
    logdiff = np.absolute(logdiff)
    if mask is not None:
        flux_interp = djs_maskinterp(flux, mask, iaxis=0)
    res = np.zeros((nTrace,len(ffiles)),dtype=flux.dtype)
    for i,f in enumerate(ffiles):
        filter_data = ascii.read(f,comment='#.*',
            names=('lam', 'respt', 'resbig', 'resnoa', 'xatm'))
        filtimg = logdiff * np.interp(newwaveimg.flatten(),filter_data['lam'].data,filter_data['respt'].data).reshape(logdiff.shape)
        if mask is not None:
            res[:,i] = (flux_interp * filtimg).sum(1)
        else:
            res[:,i] = (flux * filtimg).sum(1)
        sumfilt = filtimg.sum(1)
        res[:,i] = res[:,i] / (sumfilt + (sumfilt <= 0).astype(sumfilt.dtype))
    return res



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
#
from .. import sdss_path
from os import environ
from astropy.tests.helper import raises
from .PathTestCase import PathTestCase
#
class TestSdssPath(PathTestCase):
    """Set of tests for sdss_path().
    """
    def test_sdss_path_bad_ftype(self):
        #
        # Bad ftype
        #
        with raises(KeyError):
            p = sdss_path('fooBar',137)
    def test_sdss_path(self):
        for ftype in self.path_data:
            assert sdss_path(ftype,137,4,'301') == self.path_data[ftype]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def vactoair(vacuum):
    """Convert vacuum wavelengths to wavelengths in air.

    Parameters
    ----------
    vacuum : array-like
        Values of wavelength in vacuum in Angstroms.

    Returns
    -------
    vactoair : array-like
        Values of wavelength in air in Angstroms.

    Notes
    -----
    * Formula from Ciddor 1996  Applied Optics , 35, 1566.
    * Values of wavelength below 2000 A are not converted.
    """
    from numpy import zeros
    try:
        air = zeros(vacuum.size,dtype=vacuum.dtype)
        g = vacuum < 2000.0
    except AttributeError:
        # Most likely, vacuum is simply a float.
        air = vacuum
        g = None
        if vacuum < 2000.0:
            return air
    sigma2 = (1.0e4/vacuum)**2
    fact = 1.0 +  5.792105e-2/(238.0185 - sigma2) + 1.67917e-3/( 57.362 - sigma2)
    air = vacuum/fact
    if g is not None:
        air[g] = vacuum[g]
    return air



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_sdss_flagexist():
    from os.path import dirname, join
    from astropy.tests.helper import raises
    import pydl.pydlutils.sdss
    from ..set_maskbits import set_maskbits
    pydl.pydlutils.sdss.maskbits = set_maskbits(maskbits_file=join(dirname(__file__),'t','testMaskbits.par'))
    from .. import sdss_flagname
    names = sdss_flagname('ANCILLARY_TARGET1',2310346608843161600)
    assert tuple(names) == ('BRIGHTGAL', 'BLAZGX', 'ELG')
    names = sdss_flagname('ANCILLARY_TARGET1',2310346608843161600,concat=True)
    assert names == 'BRIGHTGAL BLAZGX ELG'
    with raises(KeyError):
        names = sdss_flagname('ABADMASK',123456789)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_unwrap_objid():
    from numpy import array
    from astropy.tests.helper import raises
    from .. import unwrap_objid
    objid = unwrap_objid(array([1237661382772195474]))
    assert objid.skyversion == 2
    assert objid.rerun == 301
    assert objid.run == 3704
    assert objid.camcol == 3
    assert objid.frame == 91
    assert objid.id == 146
    objid = unwrap_objid(array(['1237661382772195474']))
    assert objid.skyversion == 2
    assert objid.rerun == 301
    assert objid.run == 3704
    assert objid.camcol == 3
    assert objid.frame == 91
    assert objid.id == 146
    with raises(ValueError):
        objid = unwrap_objid(array([3.14159]))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_sdss_flagval():
    from os.path import dirname, join
    from astropy.tests.helper import raises
    import pydl.pydlutils.sdss
    from ..set_maskbits import set_maskbits
    pydl.pydlutils.sdss.maskbits = set_maskbits(maskbits_file=join(dirname(__file__),'t','testMaskbits.par'))
    from .. import sdss_flagval
    val = sdss_flagval('TARGET','ROSAT_A')
    assert val == 2**9
    val = sdss_flagval('ANCILLARY_TARGET1',['BLAZGX','ELG','BRIGHTGAL'])
    assert val == 2310346608843161600
    with raises(KeyError):
        val = sdss_flagval('TARGET','ROSAT_Q')
    with raises(KeyError):
        val = sdss_flagval('ABADMASK',"ABADFLAG")



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_findspec():
    """This is just a placeholder for now."""
    from .. import findspec
    # slist = findspec(infile='file.in',sdss=True)
    assert True



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def fchebyshev_split(x,m):
    """Compute the first `m` Chebyshev polynomials, but modified to allow a
    split in the baseline at ``x=0``.  The intent is to allow a model fit where
    a constant term is different for positive and negative `x`.

    Parameters
    ----------
    x : array_like
    m : int
        The number of Chebyshev polynomials to compute.  For example, if
        ``m = 4``, x > 0, T_0(x), T_1(x) and T_2(x) will be computed.

    Returns
    -------
    fchebyshev : array_like
    """
    import numpy as np
    if isinstance(x,np.ndarray):
        n = x.size
    else:
        n = 1
    if m < 2:
        raise ValueError('Order of polynomial must be at least 2.')
    try:
        dt = x.dtype
    except AttributeError:
        dt = np.float64
    leg = np.ones((m,n),dtype=dt)
    try:
        leg[0,:] = (x >= 0).astype(x.dtype)
    except AttributeError:
        leg[0,:] = np.double(x >= 0)
    if m > 2:
        leg[2,:] = x
    if m > 3:
        for k in range(3,m):
            leg[k,:] = 2.0 * x * leg[k-1,:] - leg[k-2,:]
    return leg



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_airtovac():
    from .. import airtovac
    import numpy as np
    vacuum = airtovac(1900.0)
    assert vacuum == 1900.0
    vacuum = airtovac(2000.0)
    assert np.allclose(vacuum,2000.6475)
    air = np.array([1800.0,1900.0,2000.0,2100.0,2200.0,2300.0])
    vacuum = airtovac(air)
    assert np.allclose(vacuum,np.array([1800.0,1900.0,2000.6475,2100.6664,2200.6868,2300.7083]))
    vacuum = airtovac(6056.125)
    assert np.allclose(vacuum,6057.8019)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_set_maskbits():
    from ..set_maskbits import set_maskbits
    from os.path import dirname, join
    maskbits = set_maskbits(maskbits_file=join(dirname(__file__),'t','testMaskbits.par'))
    assert (set(maskbits.keys()) ==
        set(['TARGET', 'BOSS_TARGET1', 'PRIMTARGET', 'ANCILLARY_TARGET1',
            'ZWARNING', 'TTARGET', 'SECTARGET', 'LEGACY_TARGET2',
            'LEGACY_TARGET1', 'SPECIAL_TARGET2', 'FLUXMATCH_STATUS']))
    assert (set(maskbits['TARGET'].keys()) ==
        set(['QSO_FIRST_SKIRT', 'QSO_CAP', 'GALAXY_RED', 'STAR_CARBON',
            'STAR_WHITE_DWARF', 'GALAXY_RED_II', 'GALAXY_BIG',
            'GALAXY_BRIGHT_CORE', 'SERENDIP_MANUAL', 'STAR_SUB_DWARF',
            'QSO_FIRST_CAP', 'QSO_SKIRT', 'STAR_PN', 'STAR_BHB', 'QSO_HIZ',
            'STAR_BROWN_DWARF', 'SERENDIP_FIRST', 'SOUTHERN_SURVEY',
            'STAR_RED_DWARF', 'STAR_CATY_VAR', 'QSO_REJECT', 'GALAXY',
            'SERENDIP_RED', 'SERENDIP_DISTANT', 'QSO_MAG_OUTLIER', 'ROSAT_A',
            'ROSAT_C', 'ROSAT_B', 'ROSAT_E', 'ROSAT_D', 'SERENDIP_BLUE']))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_wavevector():
    from .. import wavevector
    import numpy as np
    l = wavevector(3,5,wavemin=3,wavemax=5)
    ll = np.arange(20001,dtype='d') * 1.0e-04 + 3.0
    assert (l == ll).all()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the photoobj directory of photoop.
"""
from .sdss_calibv import sdss_calibv
from .unwrap_objid import unwrap_objid



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_is_cap_used():
    from .. import is_cap_used
    assert is_cap_used(1<<2, 2)
    assert not is_cap_used(1<<2,1)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def traceset2xy(tset,xpos=None,ignore_jump=False):
    """Convert from a trace set to an array of x,y positions.

    Parameters
    ----------
    tset : TraceSet
        A :class:`TraceSet` object.
    xpos : array-like, optional
        If provided, evaluate the trace set at these positions.  Otherwise
        the positions will be constructed from the trace set object iself.
    ignore_jump : bool, optional
        If ``True``, ignore any jump information in the `tset` object

    Returns
    -------
    traceset2xy : tuple of array-like
        The x, y positions.
    """
    return tset.xy(xpos,ignore_jump)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_sdss_objid():
    from numpy import array
    from astropy.tests.helper import raises
    from .. import sdss_objid
    assert sdss_objid(3704,3,91,146) == 1237661382772195474
    run = array([3704,1000])
    camcol = array([3,6])
    field = array([91,77])
    obj = array([146,123])
    assert (array([1237661382772195474,1237649770790322299]) == sdss_objid(run,camcol,field,obj)).all()
    #
    # Exceptions
    #
    with raises(ValueError):
        objid = sdss_objid(run,3,91,146)
    with raises(ValueError):
        objid = sdss_objid(3704,camcol,91,146)
    with raises(ValueError):
        objid = sdss_objid(3704,3,field,146)
    with raises(ValueError):
        objid = sdss_objid(3704,3,91,obj)
    with raises(ValueError):
        objid = sdss_objid(3704,3,91,146,rerun=array([137,301]))
    with raises(ValueError):
        objid = sdss_objid(3704,3,91,146,skyversion=array([2,3]))
    with raises(ValueError):
        objid = sdss_objid(3704,3,91,146,skyversion=-2)
    with raises(ValueError):
        objid = sdss_objid(3704,3,91,146,skyversion=16)
    with raises(ValueError):
        objid = sdss_objid(3704,3,91,146,rerun=-2)
    with raises(ValueError):
        objid = sdss_objid(3704,3,91,146,rerun=2**11)
    with raises(ValueError):
        objid = sdss_objid(-2,3,91,146)
    with raises(ValueError):
        objid = sdss_objid(2**16,3,91,146)
    with raises(ValueError):
        objid = sdss_objid(3704,0,91,146)
    with raises(ValueError):
        objid = sdss_objid(3704,7,91,146)
    with raises(ValueError):
        objid = sdss_objid(3704,3,-2,146)
    with raises(ValueError):
        objid = sdss_objid(3704,3,2**12,146)
    with raises(ValueError):
        objid = sdss_objid(3704,3,91,-2)
    with raises(ValueError):
        objid = sdss_objid(3704,3,91,2**16)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_cirrange():
    from .. import cirrange
    import numpy as np
    ra1 = np.linspace(-4.0*np.pi,4.0*np.pi,100)
    ra2 = cirrange(ra1,radians=True)
    assert (ra2 == (ra1 % (2.0*np.pi))).all()
    ra1 = np.rad2deg(ra1)
    ra2 = cirrange(ra1)
    assert (ra2 == (ra1 % 360.0)).all()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def unwrap_objid(objid):
    """Unwrap CAS-style objID into run, camcol, field, id, rerun.

    See pydl.pydlutils.sdss.sdss_objid() for details on how the bits
    within an objID are assigned.

    Parameters
    ----------
    objid : numpy.ndarray
        An array containing 64-bit integers or strings.  If strings are passed,
        they will be converted to integers internally.

    Returns
    -------
    unwrap_objid : numpy.recarray
        A record array with the same length as objid, with the columns
        'run', 'camcol', 'frame', 'id', 'rerun', 'skyversion'.

    Notes
    -----
    For historical reasons, the inverse of this function,
    pydl.pydlutils.sdss.sdss_objid() is not in the same namespace as this
    function.

    'frame' is used instead of 'field' because record arrays have a method
    of the same name.

    Examples
    --------
    >>> from numpy import array
    >>> from pydl.photoop.photoobj import unwrap_objid
    >>> unwrap_objid(array([1237661382772195474]))
    rec.array([(2, 301, 3704, 3, 91, 146)],
          dtype=[('skyversion', '<i4'), ('rerun', '<i4'), ('run', '<i4'), ('camcol', '<i4'), ('frame', '<i4'), ('id', '<i4')])
    """
    from numpy import bitwise_and, int64, recarray, string_, unicode_
    if objid.dtype.type is string_ or objid.dtype.type is unicode_:
        tempobjid = objid.astype(int64)
    elif objid.dtype.type is int64:
        tempobjid = objid.copy()
    else:
        raise ValueError('Unrecognized type for objid!')
    unwrap = recarray(objid.shape,dtype=[('skyversion','i4'),('rerun','i4'),('run','i4'),
        ('camcol','i4'),('frame','i4'),('id','i4')])
    unwrap.skyversion = bitwise_and(tempobjid >> 59, 2**4 - 1)
    unwrap.rerun = bitwise_and(tempobjid >> 48, 2**11 - 1)
    unwrap.run = bitwise_and(tempobjid >> 32, 2**16 - 1)
    unwrap.camcol = bitwise_and(tempobjid >> 29, 2**3 - 1)
    # unwrap.firstfield = bitwise_and(tempobjid >> 28, 2**1 - 1)
    unwrap.frame = bitwise_and(tempobjid >> 16, 2**12 - 1)
    unwrap.id = bitwise_and(tempobjid, 2**16 - 1)
    return unwrap



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def xy2traceset(xpos,ypos,**kwargs):
    """Convert from x,y positions to a trace set.

    Parameters
    ----------
    xpos, ypos : array-like
        X,Y positions corresponding as [nx,Ntrace] arrays.
    invvar : array-like, optional
        Inverse variances for fitting.
    func : str, optional
        Function type for fitting; defaults to 'legendre'.
    ncoeff : int, optional
        Number of coefficients to fit.  Defaults to 3.
    xmin, xmax : float, optional
        Explicitly set minimum and maximum values, instead of computing
        them from `xpos`.
    maxiter : int, optional
        Maximum number of rejection iterations; set to 0 for no rejection;
        default to 10.
    inmask : array-like, optional
        Mask set to 1 for good points and 0 for rejected points;
        same dimensions as `xpos`, `ypos`.  Points rejected by `inmask`
        are always rejected from the fits (the rejection is "sticky"),
        and will also be marked as rejected in the outmask attribute.
    ia, inputans, inputfunc : array-like, optional
        These arguments will be passed to :func:`func_fit`.
    xjumplo : float, optional
        x position locating start of an x discontinuity
    xjumphi : float, optional
        x position locating end of that x discontinuity
    xjumpval : float, optional
        magnitude of the discontinuity "jump" between those bounds
        (previous 3 keywords motivated by BOSS 2-phase readout)

    Returns
    -------
    xy2traceset : TraceSet
        A :class:`TraceSet` object.
    """
    from . import TraceSet
    return TraceSet(xpos,ypos,**kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def window_score(**kwargs):
    """
    For uber-resolve, score all the fields from zero to one.  If
    'rescore' is set, then write a new file 'window_flist_rescore.fits'
    rather than over-writing the file 'window_flist.fits'
    """
    # import time
    from os import environ
    from astropy.io import fits as pyfits
    from . import sdss_score
    from .. import PhotoopException
    # t0 = time.time()
    #
    # Be certain not to use global calibrations
    #
    try:
        calib_dir_save = environ['PHOTO_CALIB']
    except KeyError:
        raise PhotoopException('You have not set the environment variable PHOTO_CALIB!')
    del environ['PHOTO_CALIB']
    #
    # Read the file
    #
    try:
        resolve_dir = environ['PHOTO_RESOLVE']
    except KeyError:
        raise PhotoopException('You have not set the environment variable PHOTO_RESOLVE!')
    filename = join(resolve_dir,'window_flist.fits')
    if 'rescore' in kwargs:
        fitsmode = 'readonly'
    else:
        fitsmode = 'update'
    try:
        flist = pyfits.open(filename,mode=fitsmode)
    except IOError:
        raise PhotoopException('Unable to read FLIST file.')
    #
    # Construct the scores filling in the values to FLIST.SCORE
    #
    flist.field('SCORE')[:] = sdss_score(flist)
    if 'rescore' in kwargs:
        flist.writeto(join(resolve_dir,'window_flist_rescore.fits'))
    flist.close()
    #
    # Restore the PHOTO_CALIB variable
    #
    environ['PHOTO_CALIB'] = calib_dir_save
    # print "Elapsed time = %f sec" % (time.time()-t0,)
    return



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from numpy import allclose
from .. import TraceSet, traceset2xy, xy2traceset
from ... import PydlutilsException
from os.path import dirname,join
from astropy.io import fits
from astropy.tests.helper import raises
#
class TestTraceSet(object):
    def setup(self):
        # extracted from spFrame-b1-00057618.fits
        self.sdss = fits.open(join(dirname(__file__),'t','sdss_traceset.fits'))
        # extracted from spFrame-r1-00180406.fits
        self.boss = fits.open(join(dirname(__file__),'t','boss_traceset.fits'))
        return
    def teardown(self):
        self.sdss.close()
        self.boss.close()
        return
    def test_traceset_sdss(self):
        tset = TraceSet(self.sdss[1].data)
        assert tset.func == 'legendre'
        assert tset.coeff.shape == (320,5)
        assert not tset.has_jump
        assert tset.xRange == tset.xmax - tset.xmin
        assert tset.nx == int(tset.xmax - tset.xmin + 1)
        assert tset.xmid == 0.5 * ( tset.xmin + tset.xmax)
        x,y = traceset2xy(tset)
        tset2 = xy2traceset(x,y,ncoeff=tset.ncoeff)
        assert tset2.xmin == tset.xmin
        assert tset2.xmax == tset.xmax
        assert tset2.func == tset.func
        assert allclose(tset2.coeff,tset.coeff)
    def test_traceset_boss(self):
        tset = TraceSet(self.boss[1].data)
        assert tset.func == 'legendre'
        assert tset.coeff.shape == (500,6)
        assert tset.has_jump
        assert tset.xRange == tset.xmax - tset.xmin
        assert tset.nx == int(tset.xmax - tset.xmin + 1)
        assert tset.xmid == 0.5 * ( tset.xmin + tset.xmax)
        x,y = traceset2xy(tset)
        tset2 = xy2traceset(x,y,ncoeff=tset.ncoeff,
            xjumplo=tset.xjumplo,xjumphi=tset.xjumphi,xjumpval=tset.xjumpval)
        assert tset2.xmin == tset.xmin
        assert tset2.xmax == tset.xmax
        assert tset2.func == tset.func
        assert allclose(tset2.coeff,tset.coeff)
    def test_traceset_bad(self):
        with raises(PydlutilsException):
            tset = TraceSet(1,2,3)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_stripe_to_incl():
    from .. import stripe_to_incl
    incl = stripe_to_incl(82)
    assert incl == 0.0
    incl = stripe_to_incl(10)
    assert incl == 0.0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def read_fits_polygons(filename):
    """Read a "polygon" format FITS file.

    Unlike the IDL version of this routine, the ``XCAPS`` and ``CMPCAPS`` columns
    are *not* replaced with ``poly['CAPS']['X']`` and ``poly['CAPS']['CM']``.

    Parameters
    ----------
    filename : :class:`str`
        Name of FITS file to read.

    Returns
    -------
    read_fits_polygons : :class:`~astropy.io.fits.FITS_rec`
        The data contained in HDU 1 of the FITS file.
    """
    from astropy.io import fits
    with fits.open(filename) as hdulist:
        poly = hdulist[1].data
    # poly['CAPS'] = {'CM':list(), 'X':list()}
    # for k in range(len(poly['NCAPS'])):
    #     poly['CAPS']['CM'].append(poly['CMCAPS'][k,0:poly['NCAPS'][k]])
    #     poly['CAPS']['X'].append(poly['XCAPS'][k].reshape((8,3))[0:poly['NCAPS'][k]])
    # del poly['CMCAPS']
    # del poly['XCAPS']
    return poly



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_vactoair():
    from .. import vactoair
    import numpy as np
    air = vactoair(1900.0)
    assert air == 1900.0
    air = vactoair(2000.0)
    assert np.allclose(air,1999.3526)
    vacuum = np.array([1800.0,1900.0,2000.0,2100.0,2200.0,2300.0])
    air = vactoair(vacuum)
    assert np.allclose(air,np.array([1800.0,1900.0,1999.3526,2099.3337,2199.3133,2299.2918]))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_get_juldate():
    from .. import get_juldate
    now = get_juldate()
    assert now > 2400000.5
    assert get_juldate(0) == 40587.0 + 2400000.5
    assert get_juldate(86400) == 40588.0 + 2400000.5



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from astropy.extern import six

if six.PY3:
    # This works, but wouldn't `int` intead of `long` be OK to use here?
    # http://python3porting.com/differences.html#long
    long = int


def set_use_caps(x,cm,polygon_use_caps,add=False,tol=1.0e-10,allow_doubles=False,allow_neg_doubles=False):
    """Set the bits in use_caps for a polygon.

    Parameters
    ----------
    x : array-like
        Polygon `x` value.
    cm : array-like
        Polygon `cm` value.
    polygon_use_caps : :class:`int`
        Input value of use_caps.
    add : :class:`bool`, optional
        If ``True``, don't initialize the use_caps value to zero, use the value of
        `polygon_use_caps` instead.
    tol : :class:`float`, optional
        Tolerance used to determine whether two caps are identical.
    allow_doubles : :class:`bool`, optional
        Normally, this routine automatically sets use_caps such that no
        two caps with use_caps set are identical.
    allow_neg_doubles : :class:`bool`, optional
        Normally, two caps that are identical except for the sign of `cm`
        would be set unused.  This inhibits that behaviour.

    Returns
    -------
    set_use_caps : :class:`long`
        Value of use_caps.
    """
    import numpy as np
    from . import is_cap_used
    if add:
        use_caps = long(polygon_use_caps)
    else:
        use_caps = long(0)
    t2 = tol**2
    use_caps |= long(2)**len(cm) - long(1)
    if not allow_doubles:
        #
        # Check for doubles
        #
        for i in range(len(cm)):
            if is_cap_used(use_caps,i):
                for j in range(i+1,len(cm)):
                    if is_cap_used(use_caps,j):
                        if np.sum((x[i]-x[j])**2) < t2:
                            if ((np.absolute(cm[i]-cm[j]) < tol) or
                                ((cm[i] + cm[j]) < tol and not allow_neg_doubles)):
                                #
                                # Don't use
                                #
                                use_caps -= long(2)**j
    return use_caps



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_sdss_calib():
    from .. import sdss_calib
    foo = sdss_calib(94,6,101)
    assert foo['NMGYPERCOUNT'] == 1.0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def current_mjd():
    """Return the current MJD.
    """
    from ...goddard.astro import get_juldate
    return get_juldate() - 2400000.5



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
#
import numpy as np
import pydl.pydlutils.sdss
from astropy.tests.helper import remote_data, raises
#
# First few data lines of opBadfields.par for reference:
#
# BADFIELDS   77   focus  41  60    "Large focus offsets from -74 to 118 microns"
# BADFIELDS   77   focus 106 404    "Large focus offsets from -31 to 165 microns"
# BADFIELDS   77  astrom  30  73     "Large astrometric offset at field 39... 72"
# BADFIELDS   85  astrom   8  28     "Large astrometric offset at field 11... 27"
# BADFIELDS   85 rotator 242 253          "Large rotator offset at field 251 252"
# BADFIELDS  209  astrom   8 116      "Tel. offsets before r-band field 115 -DJS"
# BADFIELDS  209  astrom 137 175       "Tel. offsets after r-band field 145 -DJS"
# BADFIELDS  240 shutter 166 179            "parallel voltages changed frame 176"
# BADFIELDS  250  astrom 456 468               "Large astrometric offset -Manual"
# BADFIELDS  251 shutter  84 103                  "lights on in the dome -Manual"
# BADFIELDS  251   focus  97 160                   "Manual focus tests (donuts!)"
# BADFIELDS  251  astrom 147 159      "Large astrometric offset at field 156 158"
#
class TestSDSSAstrombad(object):
    def setup(self):
        self.opbadfields = np.array([
            (77, 'astrom', 30, 73, 'Large astrometric offset at field 39... 72'),
            (85, 'astrom', 8, 28, 'Large astrometric offset at field 11... 27'),
            (85, 'rotator', 242, 253, 'Large rotator offset at field 251 252'),
            (209, 'astrom', 8, 116, 'Tel. offsets before r-band field 115 -DJS'),
            (209, 'astrom', 137, 175, 'Tel. offsets after r-band field 145 -DJS'),
            (250, 'astrom', 456, 468, 'Large astrometric offset -Manual'),
            (251, 'astrom', 147, 159, 'Large astrometric offset at field 156 158')],
            dtype=[('run', '<i4'), ('problem', 'S8'), ('firstfield', '<i4'), ('lastfield', '<i4'), ('comments', 'S47')])
        pydl.pydlutils.sdss.opbadfields = self.opbadfields

    def teardown(self):
        pass

    def test_sdss_astrombad(self):
        from .. import sdss_astrombad
        assert sdss_astrombad(77,1,20) == False
        assert sdss_astrombad(77,3,35) == True
        assert sdss_astrombad(77,6,77) == False
        assert sdss_astrombad(85,1,15) == True
        assert (sdss_astrombad(np.array([77,85,251]),np.array([1,2,3]),np.array([20,15,151])) == np.array([False,True,True])).all()

    def test_sdss_astrombad_raises(self):
        from .. import sdss_astrombad
        with raises(ValueError):
            foo = sdss_astrombad(77,32,20)
        with raises(ValueError):
            foo = sdss_astrombad(-1,1,20)
        with raises(ValueError):
            foo = sdss_astrombad(2**17,1,20)
        with raises(ValueError):
            foo = sdss_astrombad(-2,1,20)
        with raises(ValueError):
            foo = sdss_astrombad(251,1,2**16)
        with raises(ValueError):
            foo = sdss_astrombad(np.array([77,85,251]),np.array([1]),np.array([20,15,151]))
        with raises(ValueError):
            foo = sdss_astrombad(np.array([77,85,251]),np.array([1,2,3]),np.array([20]))

    @remote_data
    def test_sdss_astrombad_remote(self):
        pydl.pydlutils.sdss.opbadfields = None
        from .. import sdss_astrombad
        assert sdss_astrombad(77,1,20) == False
        assert sdss_astrombad(77,3,35) == True
        assert sdss_astrombad(77,6,77) == False



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
#
def test_uniq():
    from ..uniq import uniq
    from numpy import array, sort, argsort, ones, arange
    items = array([ 1, 2, 3, 1, 5, 6, 1, 7, 3, 2, 5, 9, 11, 1 ])
    items_sorted = sort(items)
    items_argsorted = argsort(items)
    #
    # Test pre-sorted array.
    #
    u1 = uniq(items_sorted)
    equality = u1 == array([ 3,  5,  7,  9, 10, 11, 12, 13])
    assert equality.all()
    #
    # Test arg-sorted array.
    #
    u2 = uniq(items,items_argsorted)
    equality = u2 == array([ 13,  9,  8,  10, 5, 7, 11, 12])
    assert equality.all()
    assert (items_sorted[u1] == items[u2]).all()
    #
    # Test degenerate case of all identical items.
    #
    identical_items = ones((10,),dtype=items.dtype)
    u = uniq(identical_items)
    equality = u == array([9])
    assert equality.all()
    u = uniq(identical_items,arange(10,dtype=items.dtype))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def sdss_score(flist,**kwargs):
    """Score a list of imaging fields from zero to one.
    """
    import time
    import numpy as np
    import astropy.io.fits as pyfits
    from ..sdssio import sdss_name, sdss_calib
    from ...pydlutils.sdss import sdss_flagval
    if 'silent' not in kwargs:
        kwargs['silent'] = True
    lat = 32.780361
    lng = 360.0 - 105.820417
    tzone = 7
    scores = 1
    #
    # Read the PHOTO status bits
    #
    if not kwargs['silent']:
        print('Setting PHOTO status bits')
    t1 = time.time()
    nlist = flist[1].header.get('NAXIS2')
    fdata = flist[1].data
    for k in range(nlist):
        if not kwargs['silent'] and ((k % 1000) == 0):
            print("Setting PHOTO status {0:d} {1:d}".format(k, nlist))
        thisfile = sdss_name('fpFieldStat',fdata.field('RUN')[k],
            fdata.field('CAMCOL')[k],fdata.field('FIELD')[k],
            fdata.field('RERUN')[k])
        try:
            fpfield = pyfits.open(thisfile)
        except IOError:
            print("Warning: Retrying fpFieldStat file for RUN = {0:d} CAMCOL = {1:d} FIELD = {2:d} RERUN = {3}".format(fdata.field('RUN')[k],
                fdata.field('CAMCOL')[k],fdata.field('FIELD')[k],
                fdata.field('RERUN')[k]))
            try:
                fpfield = pyfits.open(thisfile)
            except IOError:
                print("Warning: Bad fpFieldStat file for RUN = {0:d} CAMCOL = {1:d} FIELD = {2:d} RERUN = {3}".format(fdata.field('RUN')[k],
                    fdata.field('CAMCOL')[k],fdata.field('FIELD')[k],
                    fdata.field('RERUN')[k]))
                fdata.field('PHOTO_STATUS')[k] = -1
                if not kwargs['silent']:
                    print('Trying tsField instead.')
                thisfile = sdss_name('tsField', fdata.field('RUN')[k],
                    fdata.field('CAMCOL')[k],fdata.field('FIELD')[k],
                    fdata.field('RERUN')[k])
                try:
                    tsfield = pyfits.open(thisfile)
                except IOError:
                    print('Warning: Bad tsField file.')
                else:
                    if not kwargs['silent']:
                        print('tsField found, using frames_status.')
                    fdata.field('PHOTO_STATUS')[k] = tsfield[1].data.field('frames_status')[0]
            else:
                fdata.field('PHOTO_STATUS')[k] = fpfield[1].data.field('status')[0]
        else:
            fdata.field('PHOTO_STATUS')[k] = fpfield[1].data.field('status')[0]
    if not kwargs['silent']:
        print("Time to set PHOTO status = {0:f} sec".format(time.time()-t1))
    #
    # Read in the PSP status
    #
    if not kwargs['silent']:
        print('Setting PSP status bits')
    t2 = time.time()
    for k in range(nlist):
        if not kwargs['silent'] and ((k % 1000) == 0):
            print("Setting PSP status {0:d} {1:d}".format(k, nlist))
        thisfile = sdss_name('psField',fdata.field('RUN')[k],
            fdata.field('CAMCOL')[k],fdata.field('FIELD')[k],
            fdata.field('RERUN')[k])
        try:
            psfield = pyfits.open(thisfile)
        except IOError:
            print("Warning: Retrying psField file for RUN = {0:d} CAMCOL = {1:d} FIELD = {2:d} RERUN = {3}".format(fdata.field('RUN')[k],
                fdata.field('CAMCOL')[k],fdata.field('FIELD')[k],
                fdata.field('RERUN')[k]))
            try:
                psfield = pyfits.open(thisfile)
            except IOError:
                print("Warning: Bad psField file for RUN = {0:d} CAMCOL = {1:d} FIELD = {2:d} RERUN = {3}".format(fdata.field('RUN')[k],
                    fdata.field('CAMCOL')[k],fdata.field('FIELD')[k],
                    fdata.field('RERUN')[k]))
                fdata.field('PSP_STATUS')[k] = -1
                fdata.field('PSF_FWHM')[k] = -1
                fdata.field('SKYFLUX')[k] = -1
        pixscale = 0.396 * np.sqrt(fdata.field('XBIN')[k]**2 + fdata.field('YBIN')[k]**2)/np.sqrt(2.0)
        calibinfo = sdss_calib(fdata.field('RUN')[k],
            fdata.field('CAMCOL')[k],fdata.field('FIELD')[k],
            fdata.field('RERUN')[k],**kwargs)
        fdata.field('PSP_STATUS')[k] = psfield[6].data.field('status')[0]
        fdata.field('PSF_FWHM')[k] = psfield[6].data.field('psf_width')[0]
        fdata.field('SKYFLUX')[k] = psfield[6].data.field('sky')[0] * calibinfo['NMGYPERCOUNT']/ pixscale**2
    if not kwargs['silent']:
        print("Time to set PSP status = {0:f} sec".format(time.time()-t2))
    #
    # Decide if each field exists in all 5 bands.
    #
    bad_bits = sdss_flagval('image_status',
        ['bad_rotator','bad_astrom','bad_focus','shutters'])
    if 'ignoreframesstatus' in kwargs:
        ignoreframesstatus = np.zeros(fdata.field('PHOTO_STATUS').shape) == 0
    else:
        ignoreframesstatus = np.zeros(fdata.field('PHOTO_STATUS').shape) == 1
    qexist = (fdata.field('PHOTO_STATUS') == 0) | ignoreframesstatus
    for k in range(5):
        qexist &= (fdata.field('IMAGE_STATUS')[:,k] & bad_bits) == 0
    #
    # Decide if each field is phtometric in all 5 bands.
    #
    unphot_bits = sdss_flagval('image_status',
        ['cloudy','unknown','ff_petals','dead_ccd','noisy_ccd'])
    qphot = fdata.field('SUN_ANGLE') < -12
    for k in range(5):
        qphot &= (fdata.field('IMAGE_STATUS')[:,k] & unphot_bits) == 0
    for k in range(5):
        qphot &= (((fdata.field('PSP_STATUS')[:,k] & 31) <= 2) |
            (fdata.field('XBIN') > 1) | ignoreframesstatus)
    #
    # Now set the score for each field
    #
    sensitivity = (0.7 / (fdata.field('PSF_FWHM')[:,2] * np.sqrt(fdata.field('SKYFLUX')[:,2]))) < 0.4
    fdata.field('SCORE')[:] = qexist * (0.1 + 0.5*qphot + sensitivity)
    ibinned = np.find(fdata.field('XBIN') > 1)
    if len(ibinned) > 0:
        fdata.field('SCORE')[ibinned] *= 0.1
    #
    # Look for any NaN values, which could happen for example if there
    # is a corrupted psField file and PSF_FWHM or SKYFLUX was negative.
    #
    ibad = np.find(~np.isfinite(fdata.field('SCORE')))
    if len(ibad) > 0:
        print("Warning: Changing NaN scores for {0:d} fields to zero.".format(len(ibad)))
        fdata.field('SCORE')[ibad] = 0
    return fdata.field('SCORE')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_radec_to_munu():
    from astropy import units as u
    from astropy.coordinates import ICRS
    from .. import SDSSMuNu
    radec = ICRS(ra=0.0*u.deg,dec=0.0*u.deg)
    munu = radec.transform_to(SDSSMuNu(stripe=10))
    assert munu.mu.value == 0.0
    assert munu.nu.value == 0.0
    assert munu.incl.value == 0.0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_sdss_calibv():
    from numpy.testing import assert_allclose
    from .. import sdss_calibv
    assert_allclose(0.2650306748466258,sdss_calibv())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
#
from astropy.coordinates import Angle, frame_transform_graph, FunctionTransform, ICRS
from . import SDSSMuNu
#
@frame_transform_graph.transform(FunctionTransform, ICRS, SDSSMuNu)
def radec_to_munu(icrs_frame,munu):
    """Convert from equatorial coordinates to SDSS great circle coordinates.

    Parameters
    ----------
    icrs_frame : :class:`~astropy.coordinates.ICRS`
        Equatorial coordinates (RA, Dec).

    Returns
    -------
    radec_to_munu : :class:`~pydl.pydlutils.coord.SDSSMuNu`
        SDSS great circle coordinates (mu, nu).
    """
    from astropy import units as u
    import numpy as np
    # from pydlutils.coord import stripe_to_eta
    # from pydlutils.goddard.misc import cirrange
    # if 'stripe' in kwargs:
    #     node = 95.0
    #     incl = stripe_to_incl(kwargs['stripe'])
    # elif 'node' in kwargs and 'incl' in kwargs:
    #     node = kwargs['node']
    #     incl = kwargs['incl']
    # else:
    #     raise ValueError('Must specify either STRIPE or NODE,INCL!')
    # if ra.size != dec.size:
    #     raise ValueError('Number of elements in RA and DEC must agree!')
    sinra = np.sin((icrs_frame.ra - munu.node).to(u.radian).value)
    cosra = np.cos((icrs_frame.ra - munu.node).to(u.radian).value)
    sindec = np.sin(icrs_frame.dec.to(u.radian).value)
    cosdec = np.cos(icrs_frame.dec.to(u.radian).value)
    sini = np.sin(munu.incl.to(u.radian).value)
    cosi = np.cos(munu.incl.to(u.radian).value)
    x1 = cosdec * cosra
    y1 = cosdec * sinra
    z1 = sindec
    x2 = x1
    y2 = y1 * cosi + z1 * sini
    z2 = -y1 * sini + z1 * cosi
    mu = Angle(np.arctan2(y2,x2),unit=u.radian) + munu.node
    nu = Angle(np.arcsin(z2),unit=u.radian)
    # if 'phi' in kwargs:
    #     sinnu = np.sin(np.deg2rad(nu))
    #     cosnu = np.cos(np.deg2rad(nu))
    #     sinmu = np.sin(np.deg2rad(mu-node))
    #     cosmu = np.cos(np.deg2rad(mu-node))
    #     phi = np.rad2deg(np.arctan2(cosmu * sini,
    #         (-sinmu * sinnu * sini + cosnu * cosi)*cosnu))
    #     return (ra,dec,phi)
    # else:
    #     return (ra,dec)
    return SDSSMuNu(mu=mu,nu=nu,stripe=munu.stripe)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_munu_to_radec():
    from astropy import units as u
    from astropy.coordinates import ICRS
    from .. import SDSSMuNu
    munu = SDSSMuNu(mu=0.0*u.deg,nu=0.0*u.deg,stripe=10)
    radec = munu.transform_to(ICRS)
    assert radec.ra.value == 0.0
    assert radec.dec.value == 0.0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_fchebyshev():
    from .. import fchebyshev
    import numpy as np
    from astropy.tests.helper import raises
    x = np.array([-1,-0.5,0,0.5,1],dtype='d')
    #
    # Test order
    #
    with raises(ValueError):
        f = fchebyshev(x,0)
    #
    # m = 1
    #
    f = fchebyshev(x,1)
    assert (f == np.ones((1,x.size),dtype='d')).all()
    #
    # m = 2
    #
    f = fchebyshev(x,2)
    foo = np.ones((2,x.size),dtype='d')
    foo[1,:] = x
    assert np.allclose(f,foo)
    #
    # m = 3
    #
    f = fchebyshev(x,3)
    foo = np.ones((3,x.size),dtype='d')
    foo[1,:] = x
    foo[2,:] = (2.0*x**2 - 1.0)
    assert np.allclose(f,foo)
    #
    # m = 4
    #
    f = fchebyshev(x,4)
    foo = np.ones((4,x.size),dtype='d')
    foo[1,:] = x
    foo[2,:] = (2.0*x**2 - 1.0)
    foo[3,:] = (4.0*x**3 - 3.0*x)
    assert np.allclose(f,foo)
    #
    # random float
    #
    f = fchebyshev(2.88,3)
    assert np.allclose(f,np.array([[1.00], [2.88], [15.5888]]))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def sdss_calib(run,camcol,field,rerun='',**kwargs):
    """Read photometric calibration solutions from calibPhotom or calibPhotomGlobal files.

    Parameters
    ----------
    run : int
        Photo run number
    camcol : int
        Camcol number
    field : int
        Field number
    rerun : str, optional
        Photometric reduction number, as a string.

    Returns
    -------
    sdss_calib : dict
        A dictionary containing the 'NMGYPERCOUNT' keyword.

    Notes
    -----
    Currently, this is just a placeholder.
    """
    return { 'NMGYPERCOUNT':1.0 }



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def airtovac(air):
    """Convert air wavelengths to wavelengths in vacuum.

    Parameters
    ----------
    air : array-like
        Values of wavelength in air in Angstroms.

    Returns
    -------
    airtovac : array-like
        Values of wavelength in vacuum in Angstroms.

    Notes
    -----
    * Formula from Ciddor 1996  Applied Optics , 35, 1566.
    * Values of wavelength below 2000 A are not converted.
    """
    from numpy import zeros
    try:
        vacuum = zeros(air.size,dtype=air.dtype) + air
        g = vacuum < 2000.0
    except AttributeError:
        # Most likely, vacuum is simply a float.
        vacuum = air
        g = None
        if air < 2000.0:
            return air
    for k in range(2):
        sigma2 = (1.0e4/vacuum)**2
        fact = 1.0 +  5.792105e-2/(238.0185 - sigma2) + 1.67917e-3/( 57.362 - sigma2)
        vacuum = air * fact
    if g is not None:
        vacuum[g] = air[g]
    return vacuum



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the coord directory of idlutils.
"""
from .SDSSMuNu import SDSSMuNu
from .current_mjd import current_mjd
from .munu_to_radec import munu_to_radec
from .radec_to_munu import radec_to_munu
from .stripe_to_eta import stripe_to_eta
from .stripe_to_incl import stripe_to_incl



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def wavevector(minfullwave,maxfullwave,zeropoint=3.5,binsz=1.0e-4,wavemin=None,wavemax=None):
    """Return an array of wavelengths.

    Parameters
    ----------
    minfullwave : float
        Minimum wavelength.
    maxfullwave : float
        Maximum wavelength.
    zeropoint : float, optional
        Offset of the input wavelength values.
    binsz : float, optional
        Separation between wavelength values.
    wavemin : float, optional
        If this is set the values of `minfullwave` and `zeropoint` are ignored.
    wavemax : float, optional
        If this is set the value of `maxfullwave` is ignored.

    Returns
    -------
    wavevector : numpy.ndarray
        Depending on the values of `minfullwave`, `binsz`, etc., the resulting
        array could be interpreted as an array of wavelengths or an array of
        log(wavelength).
    """
    from numpy import arange
    if wavemin is not None:
        spotmin = 0
        if wavemax is not None:
            spotmax = int((wavemax - wavemin)/binsz)
        else:
            spotmax = int((maxfullwave - wavemin)/binsz)
            wavemax = spotmax * binsz + wavemin
    else:
        spotmin = int((minfullwave - zeropoint)/binsz) + 1
        spotmax = int((maxfullwave - zeropoint)/binsz)
        wavemin = spotmin * binsz + zeropoint
        wavemax = spotmax * binsz + zeropoint
    nfinalpix = spotmax - spotmin + 1
    finalwave = arange(nfinalpix,dtype='d') *binsz + wavemin
    return finalwave



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_sdssflux2ab():
    import numpy as np
    from .. import sdssflux2ab
    correction = np.array([-0.042, 0.036, 0.015, 0.013, -0.002])
    mags = np.zeros((2,5),dtype='d')
    mags[0,:] = 18.0
    mags[1,:] = 19.0
    ab = sdssflux2ab(mags,magnitude=True)
    assert (ab == (mags + correction)).all()
    flux = 10**((22.5 - mags)/2.5) # nanomaggies
    ab = sdssflux2ab(flux)
    assert np.allclose(ab,10**((22.5 - (mags + correction))/2.5))
    ivar = 1.0/flux
    ab = sdssflux2ab(ivar,ivar=True)
    assert np.allclose(ab,ivar/(10**(-2.0*correction/2.5)))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def fix_medfilt(array,width):
    """Wrap medfilt so that the results more closely resemble IDL MEDIAN().
    """
    from numpy import arange
    from scipy.signal import medfilt
    medarray = medfilt(array,min(width,array.size))
    istart = int((width-1)/2)
    iend = array.size - int((width+1)/2)
    i = arange(array.size)
    w = (i < istart) | (i > iend)
    medarray[w] = array[w]
    return medarray
#
#
#
def fix_medfilt2d(array,width):
    """Wrap medfilt2d so that the results more closely resemble IDL MEDIAN().
    """
    from numpy import arange
    from scipy.signal import medfilt2d
    medarray = medfilt2d(array,min(width,array.size))
    istart = int((width-1)/2)
    iend = (array.shape[0] - int((width+1)/2), array.shape[1] - int((width+1)/2))
    i = np.arange(array.shape[0])
    j = np.arange(array.shape[1])
    w = ((i < istart) | (i > iend[0]), (j < istart) | (j > iend[1]))
    medarray[w[0],w[1]] = array[w[0],w[1]]
    return medarray
#
#
#
def djs_median(array,dimension=None,width=None,boundary='none'):
    """Compute the median of an array.

    Use a filtering box or collapse the image along one dimension.

    Parameters
    ----------
    array : :class:`numpy.ndarray`
        input array
    dimension : :class:`int`, optional
        Compute the median over this dimension. It is an error to specify both
        `dimension` and `width`.
    width : :class:`int`, optional
        Width of the median window. In general, this should be an odd
        integer.  It is an error to specify both `dimension` and `width`.
    boundary : { 'none', 'reflect', 'nearest', 'wrap' }, optional
        Boundary condition to impose.  'none' means no filtering is done within
        `width`/2 of the boundary.  'reflect' means reflect pixel values around the
        boundary. 'nearest' means use the values of the nearest boundary pixel.
        'wrap' means wrap pixel values around the boundary. 'nearest' and 'wrap'
        are not implemented.

    Returns
    -------
    djs_median : :class:`numpy.ndarray`
        The output.  If neither `dimension` nor `width` are set, this is a scalar
        value, just the output of ``numpy.median()``.  If `dimension` is set,
        then the result simply ``numpy.median(array,dimension)``.
        If `width` is set, the result has the same shape as the input array.
    """
    import numpy as np
    if dimension is None and width is None:
        return np.median(array)
    elif width is None:
        return np.median(array,axis=dimension)
    elif dimension is None:
        if width == 1:
            return array
        if boundary == 'none':
            if array.ndim == 1:
                return fix_medfilt(array,width)
            elif array.ndim == 2:
                return fix_medfilt2d(array,width)
            else:
                raise ValueError('Unsupported number of dimensions with this boundary condition.')
        elif boundary == 'reflect':
            padsize = int(np.ceil(width/2.0))
            if array.ndim == 1:
                bigarr = np.zeros(array.shape[0]+2*padsize,dtype=array.dtype)
                bigarr[padsize:padsize+array.shape[0]] = array
                bigarr[0:padsize] = array[0:padsize][::-1]
                bigarr[padsize+array.shape[0]:padsize*2+array.shape[0]] = array[array.shape[0]-padsize:array.shape[0]][::-1]
                f = fix_medfilt(bigarr,width)
                medarray = f[padsize:padsize+array.shape[0]]
                return medarray
            elif array.ndim == 2:
                bigarr = np.zeros((array.shape[0]+2*padsize,array.shape[1]+2*padsize),dtype=array.dtype)
                bigarr[padsize:padsize+array.shape[0],padsize:padsize+array.shape[1]] = array
                # Copy into top + bottom
                bigarr[0:padsize,padsize:array.shape[1]+padsize] = array[0:padsize,:][::-1,:]
                bigarr[array.shape[0]+padsize:bigarr.shape[0],padsize:array.shape[1]+padsize] = array[array.shape[0]-padsize:array.shape[0],:][::-1,:]
                # Copy into left + right
                bigarr[padsize:array.shape[0]+padsize,0:padsize] = array[:,0:padsize][:,::-1]
                bigarr[padsize:array.shape[0]+padsize] = array[:array.shape[1]-padsize:array.shape[1]][:,::-1]
                # Copy into top left
                bigarr[0:padsize,0:padsize] = foo[0:padsize,0:padsize][::-1,::-1]
                # Copy into top right
                bigarr[0:padsize,bigarr.shape[1]-padsize:bigarr.shape[1]] = array[0:padsize,array.shape[1]-padsize:array.shape[1]][::-1,::-1]
                # Copy into bottom left
                bigarr[bigarr.shape[0]-padsize:bigarr.shape[0],0:padsize] = array[array.shape[0]-padsize,0:padsize][::-1,::-1]
                # Copy into bottom right
                bigarr[bigarr.shape[0]-padsize:bigarr.shape[0],bigarr.shape[1]-padsize:bigarr.shape[1]] = array[ar]
                f = fix_medfilt2d(bigarr,min(width,array.size))
                medarray = f[padsize:array.shape[0]+padsize,padsize:array.shape[1]+padsize]
                return medarray
            else:
                raise ValueError('Unsupported number of dimensions with this boundary condition.')
        elif boundary == 'nearest':
            raise ValueError('This boundary condition not implemented')
        elif boundary == 'wrap':
            raise ValueError('This boundary condition not implemented')
        else:
            raise ValueError('Unknown boundary condition.')
    else:
        raise ValueError('Invalid to specify both dimension & width.')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage corresponds to the trace directory of idlutils.
"""
from .fchebyshev import fchebyshev
from .fchebyshev_split import fchebyshev_split
from .fpoly import fpoly
from .func_fit import func_fit
from .TraceSet import TraceSet
from .traceset2xy import traceset2xy
from .xy2traceset import xy2traceset



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
#
from astropy.coordinates import Angle, frame_transform_graph, FunctionTransform, ICRS
from . import SDSSMuNu
#
@frame_transform_graph.transform(FunctionTransform, SDSSMuNu, ICRS)
def munu_to_radec(munu,icrs_frame):
    """Convert from SDSS great circle coordinates to equatorial coordinates.

    Parameters
    ----------
    munu : :class:`~pydl.pydlutils.coord.SDSSMuNu`
        SDSS great circle coordinates (mu, nu).

    Returns
    -------
    munu_to_radec : :class:`~astropy.coordinates.ICRS`
        Equatorial coordinates (RA, Dec).
    """
    from astropy import units as u
    import numpy as np
    # from pydlutils.coord import stripe_to_eta
    # from pydlutils.goddard.misc import cirrange
    # if 'stripe' in kwargs:
    #     node = 95.0
    #     incl = stripe_to_incl(kwargs['stripe'])
    # elif 'node' in kwargs and 'incl' in kwargs:
    #     node = kwargs['node']
    #     incl = kwargs['incl']
    # else:
    #     raise ValueError('Must specify either STRIPE or NODE,INCL!')
    # if mu.size != nu.size:
    #     raise ValueError('Number of elements in MU and NU must agree!')
    sinnu = np.sin(munu.nu.to(u.radian).value)
    cosnu = np.cos(munu.nu.to(u.radian).value)
    sini = np.sin(munu.incl.to(u.radian).value)
    cosi = np.cos(munu.incl.to(u.radian).value)
    sinmu = np.sin((munu.mu - munu.node).to(u.radian).value)
    cosmu = np.cos((munu.mu - munu.node).to(u.radian).value)
    xx = cosmu * cosnu
    yy = sinmu * cosnu * cosi - sinnu * sini
    zz = sinmu * cosnu * sini + sinnu * cosi
    ra = Angle(np.arctan2(yy,xx),unit=u.radian) + munu.node
    dec = Angle(np.arcsin(zz),unit=u.radian)
    # if 'phi' in kwargs:
    #     phi = np.rad2deg(np.arctan2(cosmu * sini,
    #         (-sinmu * sinnu * sini + cosnu * cosi)*cosnu))
    #     return (ra,dec,phi)
    # else:
    #     return (ra,dec)
    return ICRS(ra=ra,dec=dec).transform_to(icrs_frame)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def stripe_to_incl(stripe):
    """Convert from SDSS stripe number to an inclination relative to the equator.

    Parameters
    ----------
    stripe : :class:`int` or :class:`numpy.ndarray`
        SDSS Stripe number.

    Returns
    -------
    stripe_to_incl : :class:`float` or :class:`numpy.ndarray`
        Inclination of the stripe relative to the equator (Dec = 0).
    """
    from . import stripe_to_eta
    dec_center = 32.5
    eta_center = stripe_to_eta(stripe)
    incl = eta_center + dec_center
    return incl



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_func_fit():
    from .. import func_fit
    import numpy as np
    from astropy.tests.helper import raises
    np.random.seed(137)
    x = np.linspace(-5,5,50)
    y = x**2 + 2*x + 1 + 0.05*np.random.randn(50)
    #
    # Bad inputs
    #
    with raises(ValueError):
        foo = func_fit(x,np.array([1,2,3]),3)
    with raises(ValueError):
        foo = func_fit(x,y,3,invvar=np.array([0,0,0]))
    with raises(ValueError):
        foo = func_fit(x,y,3,inputfunc=np.array([1.0,1.0,1.0]))
    with raises(KeyError):
        foo = func_fit(x,y,3,function_name='npoly')
    #
    # No good points
    #
    invvar = np.zeros(x.shape,dtype=x.dtype)
    res,yfit = func_fit(x,y,3,invvar=invvar)
    assert (res == np.zeros((3,),dtype=x.dtype)).all()
    assert (yfit == 0*x).all()
    #
    # One good point
    #
    invvar[2] = 1.0
    res,yfit = func_fit(x,y,3,invvar=invvar)
    assert (invvar > 0).nonzero()[0][0] == 2
    assert res[0] == y[2]
    assert (yfit == y[2]).all()
    #
    # Various points
    #
    invvar = 1.0/(np.random.random(x.shape)**2)
    assert (invvar < 2).any()
    invvar[invvar < 2] = 0
    res,yfit = func_fit(x,y,3,invvar=invvar,function_name='poly')
    # assert np.allclose(res,np.array([ 0.99665423,  1.9945388 ,  1.00172303]))
    assert np.allclose(res,np.array([ 0.99996197,  1.99340315,  1.00148004]))
    #
    # Fixed points
    #
    res,yfit = func_fit(x,y,3,invvar=invvar,function_name='poly',ia=np.array([False,True,True]),inputans=np.array([1.0,0,0]))
    # assert np.allclose(res,np.array([ 1.        ,  1.99454782,  1.00149949]))
    assert np.allclose(res,np.array([ 1.        ,  1.99340359,  1.00147743]))
    res,yfit = func_fit(x,y,3,invvar=invvar,function_name='poly',ia=np.array([False,True,False]),inputans=np.array([1.0,0,1.0]))
    # assert np.allclose(res,np.array([ 1.        ,  1.99403239,  1.        ]))
    assert np.allclose(res,np.array([ 1.        ,  1.99735654,  1.        ]))
    #
    # inputfunc
    #
    res,yfit = func_fit(x,y,3,invvar=invvar,function_name='poly',inputfunc=np.ones(x.shape,dtype=x.dtype))
    # assert np.allclose(res,np.array([ 0.99665423,  1.9945388 ,  1.00172303]))
    assert np.allclose(res,np.array([ 0.99996197,  1.99340315,  1.00148004]))
    #
    # Generate inputans
    #
    y = x**2 + 2*x + 0.05*np.random.randn(50)
    res,yfit = func_fit(x,y,3,invvar=invvar,function_name='poly',ia=np.array([False,True,True]))
    assert np.allclose(res,np.array([ 0.        ,  1.99994188,  0.99915111]))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def stripe_to_eta(stripe):
    """Convert from SDSS great circle coordinates to equatorial coordinates.

    Parameters
    ----------
    stripe : :class:`int` or :class:`numpy.ndarray`
        SDSS Stripe number.

    Returns
    -------
    stripe_to_eta : :class:`float` or :class:`numpy.ndarray`
        The eta value in the SDSS (lambda,eta) coordinate system.
    """
    stripe_sep = 2.5
    eta = stripe * stripe_sep - 57.5
    if stripe > 46:
        eta -= 180.0
    return eta



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_djs_laxisnum():
    import numpy as np
    from astropy.tests.helper import raises
    from .. import djs_laxisnum
    #
    # 1d
    #
    assert (np.zeros((4,),dtype='i4') == djs_laxisnum((4,))).all()
    #
    # 2d
    #
    l = np.array([[0,0,0,0],[1,1,1,1],[2,2,2,2],[3,3,3,3]],dtype='i4')
    assert (l == djs_laxisnum((4,4))).all()
    assert (l.T == djs_laxisnum((4,4),iaxis=1)).all()
    with raises(ValueError):
        foo = djs_laxisnum((4,4),iaxis=2)
    #
    # 3d
    #
    l = np.zeros((4,4,4),dtype='i4')
    l[1,:,:] = 1
    l[2,:,:] = 2
    l[3,:,:] = 3
    assert (l == djs_laxisnum((4,4,4))).all()
    assert (l.swapaxes(0,1) == djs_laxisnum((4,4,4),iaxis=1)).all()
    assert (l.swapaxes(0,2) == djs_laxisnum((4,4,4),iaxis=2)).all()
    with raises(ValueError):
        foo = djs_laxisnum((4,4,4),iaxis=3)
    #
    # More d
    #
    with raises(ValueError):
        foo = djs_laxisnum((4,4,4,4))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def test_sdss_flagexist():
    from os.path import dirname, join
    import pydl.pydlutils.sdss
    from ..set_maskbits import set_maskbits
    pydl.pydlutils.sdss.maskbits = set_maskbits(maskbits_file=join(dirname(__file__),'t','testMaskbits.par'))
    from .. import sdss_flagexist
    assert sdss_flagexist('TARGET','ROSAT_A')
    assert sdss_flagexist('TARGET',['rosat_a','rosat_b','rosat_c','rosat_d'])
    l, f = sdss_flagexist('target','FOOBAR',flagexist=True)
    assert not l
    assert f
    l, which = sdss_flagexist('TARGET',['rosat_a','rosat_b','rosat_c','rosat_d','foobar'],whichexist=True)
    assert not l
    assert tuple(which) == (True,True,True,True,False)
    l, f, which = sdss_flagexist('TARGET',['rosat_a','rosat_b','rosat_c','rosat_d','foobar'],
        flagexist=True, whichexist=True)
    assert not l
    assert f
    assert tuple(which) == (True,True,True,True,False)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
from __future__ import print_function
#
#
#
def get_juldate(seconds=None):
    """Returns the current Julian date.

    Uses the MJD trick & adds the offset to get JD.

    Parameters
    ----------
    seconds : int or float, optional
        Time in seconds since the UNIX epoch.  This should only be used
        for testing.

    Returns
    -------
    get_juldate : float
        The Julian Day number as a floating point number.

    Notes
    -----
    Do not use this function if high precision is required, or if you are
    concerned about the distinction between UTC & TAI.
    """
    from time import time
    if seconds is None:
        t = time()
    else:
        t = seconds
    mjd = t/86400.0 + 40587.0
    return mjd + 2400000.5
#
#
#
def get_juldate_main(): # pragma: no cover
    """Allow this module to be run in scripts.
    """
    jd = get_juldate()
    print(jd)
    return 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
def flegendre(x,m):
    """Compute the first `m` Legendre polynomials.

    Parameters
    ----------
    x : array_like
    m : int
        The number of Legendre polynomials to compute.  For example, if
        ``m = 3``, P_0(x), P_1(x) and P_2(x) will be computed.

    Returns
    -------
    flegendre : array_like
    """
    import numpy as np
    from scipy.special import legendre
    if isinstance(x,np.ndarray):
        n = x.size
    else:
        n = 1
    if m < 1:
        raise ValueError('Number of Legendre polynomials must be at least 1.')
    try:
        dt = x.dtype
    except AttributeError:
        dt = np.float64
    leg = np.ones((m,n),dtype=dt)
    if m >= 2:
        leg[1,:] = x
    if m >= 3:
        for k in range(2,m):
            leg[k,:] = np.polyval(legendre(k),x)
    return leg
